
import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.nio.*;
import java.nio.charset.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.text.html.HTML.*;

class FileReader extends InputStreamReader {
    FileReader(File file) throws IOException {
        this(file, null);
    }
    FileReader(File file, String encoding) throws IOException {
        super(new BufferedInputStream(new FileInputStream(file)), decoder(encoding));
    }
    
    private static CharsetDecoder decoder(String encoding) {
        Charset charset = (encoding == null) ? Charset.defaultCharset() : Charset.forName(encoding);
        CharsetDecoder decoder = charset.newDecoder();
        decoder.onMalformedInput(CodingErrorAction.REPORT);
        decoder.onUnmappableCharacter(CodingErrorAction.REPORT);
        return decoder;
    }
    
}
class FileWriter extends OutputStreamWriter {
    FileWriter(File file) throws IOException {
        super(new BufferedOutputStream(new FileOutputStream(file)));
    }
    FileWriter(File file, String encoding) throws IOException {
        super(new BufferedOutputStream(new FileOutputStream(file)), encoding);
    }
}

public class Main {
    private final String shortName;
    private final String title;
    private final File inputDir;
    private final File outputDir;
    private final String encodings;
    private final PrintWriter help;
    private final PrintWriter index;
    private final PrintWriter toc;
    public Main(String shortName, String title, File inputDir, File outputDir, String encodings) throws Exception {
        this.shortName = shortName;
        this.title = title;
        this.inputDir = inputDir;
        this.outputDir = outputDir;
        this.encodings = encodings;
        
        outputDir.mkdirs();
        help = new PrintWriter(new FileWriter(new File(outputDir, "winhelp.hhp")));
        index = new PrintWriter(new FileWriter(new File(outputDir, "index.hhk")));
        toc = new PrintWriter(new FileWriter(new File(outputDir, "toc.hhc")));
        
        go();
        
        help.close();
        index.close();
        toc.close();
        
    }
    private void go() throws Exception {
        help.println("[OPTIONS]");
        //help.println("Compatibility=1.1 or later");
        help.println("Compiled file=" + shortName + ".chm");
        help.println("Contents file=toc.hhc");
        help.println("Default Window=Main");
        help.println("Default topic=index.html");
        //help.println("Display compile progress=No");
        help.println("Full-text search=Yes");
        help.println("Index file=index.hhk");
        //help.println("Language=0x409"); // U.S. English
        //help.println("Language=0x411"); // Japanese
        help.println("Title=" + title);
        
        help.println();
        help.println("[WINDOWS]");
        help.println("Main=\"" + toHTML(title) + "\",\"toc.hhc\",\"index.hhk\",\"index.html\",,,,,,0x63520,,0x385e,,,,,,,,0");
        
        help.println();
        help.println("[FILES]");
        //listFiles(new File("."));
        
        toc.println("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">");
        toc.println("<HTML>");
        toc.println("<HEAD>");
        //toc.println("<meta name=\"GENERATOR\" content=\"Microsoft&reg; HTML Help Workshop 4.1\">");
        toc.println("<!-- Sitemap 1.0 -->");
        toc.println("</HEAD><BODY>");
        toc.println("<OBJECT type=\"text/site properties\">");
        toc.println("<param name=\"ImageType\" value=\"Folder\">");
        toc.println("</OBJECT>");
        
        File indexFile = new File(inputDir, "index.htm");
        if (!indexFile.exists())
            indexFile = new File(inputDir, "index.html");
        if (indexFile.exists()) {
            toc.println("<UL>");
            help.println(toHTML(indexFile.getName()));
            buildTOC(indexFile);
            toc.println("</UL>");
        }
        toc.println("</BODY></HTML>");
        
        index.println("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">");
        index.println("<HTML>");
        index.println("<HEAD>");
        //index.println("<meta name=\"GENERATOR\" content=\"Microsoft&reg; HTML Help Workshop 4.1\">");
        index.println("<!-- Sitemap 1.0 -->");
        index.println("</HEAD><BODY>");
        index.println("<UL>");
        
        Map<String, File> firstCharToTemporaryFile = new TreeMap<String, File>();
        Map<String, DataOutputStream> firstCharToTemporaryFileStream = new LinkedHashMap<String, DataOutputStream>() {
            protected boolean removeEldestEntry(Map.Entry<String, DataOutputStream> eldest) {
                if (size() > 10) {
                    OutputStream stream = eldest.getValue();
                    try {
                        stream.close();
                    } catch (IOException x) {
                    }
                    return true;
                }
                return false;
            }
        };
        buildIndex(firstCharToTemporaryFile, firstCharToTemporaryFileStream, inputDir);
        for (OutputStream stream : firstCharToTemporaryFileStream.values()) {
            stream.close();
        }
        for (File temporaryFile : firstCharToTemporaryFile.values()) {
            Map<String, SortedSet<String>> map = new TreeMap<String, SortedSet<String>>();
            DataInputStream dis = new DataInputStream(new BufferedInputStream(new FileInputStream(temporaryFile)));
            for (;;) {
                try {
                    String key = dis.readUTF();
                    String value = dis.readUTF();
                    SortedSet<String> values = map.get(key);
                    if (values == null) {
                        values = new TreeSet<String>();
                        map.put(key, values);
                    }
                    values.add(value);
                } catch (EOFException x) {
                    break;
                }
            }
            dis.close();
            for (Map.Entry<String, SortedSet<String>> entry : map.entrySet()) {
                outputIndex(entry.getKey(), entry.getValue());
            }
        }
        
        index.println("</UL>");
        index.println("</BODY></HTML>");
        
        copyFiles(0, inputDir, outputDir);
        
    }
    /*
    private static Set ignoredFiles = new HashSet(Arrays.asList(new String[] { "winhelp.hhp", "index.hhk", "toc.hhc" }));
    private void listFiles(File dir) {
        File[] files = dir.listFiles();
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                if (files[i].isFile() && !ignoredFiles.contains(files[i].getName()))
                    help.println(toHTML(files[i].toString().substring(2)));
            }
            for (int i = 0; i < files.length; i++) {
                if (files[i].isDirectory())
                    listFiles(files[i]);
            }
        }
    }
    */
    
    private void buildTOC(File indexFile) throws Exception {
        if (indexFile.exists()) {
            toc.println("<LI><OBJECT type=\"text/sitemap\">");
            toc.println("  <param name=\"Name\" value=\"" + getTitle(indexFile) + "\">");
            toc.println("  <param name=\"Local\" value=\"" + toHTML(indexFile.toString().substring(inputDir.toString().length() + 1)) + "\">");
            toc.println("</OBJECT>");
        }
        boolean hasChild = false;
        
        File[] files = indexFile.getParentFile().listFiles();
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                if (files[i].isDirectory()) {
                    File file = new File(files[i], "index.htm");
                    if (!file.exists())
                        file = new File(files[i], "index.html");
                    if (file.exists()) {
                        if (!hasChild) {
                            toc.println("<UL>");
                            hasChild = true;
                        }
                        buildTOC(file);
                    }
                }
            }
            if (hasChild)
                toc.println("</UL>");
        }
    }
    
    // ここの実装は適当。
    private final Pattern TITLE_PATTERN = Pattern.compile("<TITLE>(([^<]|[\n\r])*)</TITLE>", Pattern.CASE_INSENSITIVE | Pattern.MULTILINE);
    private String getTitle(File file) throws Exception {
        Matcher m = TITLE_PATTERN.matcher(Decoder.decode(file, encodings));
        if (m.find()) {
            return m.group(1).trim();
        }
        return "";
    }
    
    // 以下ではタイトルは取得できない。
    //private String getTitle(File file) throws Exception {
    //    StringBuilder result = new StringBuilder();
    //    
    //    HTMLDocument doc = getHTMLDocument(file);
    //    Element[] elements = doc.getRootElements();
    //    for (int i = 0; i < elements.length; i++) {
    //        pickTitle(result, elements[i]);
    //    }
    //    
    //    return result.toString();
    //}
    //private void pickTitle(StringBuilder result, Element element) throws Exception {
    //    if (!element.isLeaf()) {
    //        int count = element.getElementCount();
    //        for (int j = 0; j < count; j++) {
    //            Element e = element.getElement(j);
    //            Tag tag = HTML.getTag(e.getName());
    //            if (tag == Tag.TITLE) {
    //                int start = e.getStartOffset();
    //                int end = e.getEndOffset();
    //                String text = e.getDocument().getText(start, end - start);
    //                result.append(text);
    //            } else {
    //                pickTitle(result, e);
    //            }
    //        }
    //    }
    //}
    
    private void buildIndex(Map<String, File> firstCharToTemporaryFile, Map<String, DataOutputStream> firstCharToTemporaryFileStream, File dir) throws Exception {
        File[] files = dir.listFiles();
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                if (files[i].isFile()) {
                    String name = files[i].getName();
                    String head = "index-";
                    String tail = ".html";
                    if (/*name.equals("index-1.html") &&*/ name.startsWith(head) && name.endsWith(tail)) {
                        System.out.println("indexing " + files[i]);
                        String mid = name.substring(head.length(), name.length() - tail.length());
                        boolean hit;
                        try {
                            Integer.parseInt(mid);
                            hit = true;
                        } catch (NumberFormatException x) {
                            hit = mid.equals("all");
                        }
                        if (hit) {
                            help.println(toHTML(files[i].toString().substring(inputDir.toString().length() + 1)));
                            buildIndexFromFile(firstCharToTemporaryFile, firstCharToTemporaryFileStream, files[i]);
                        }
                    }
                }
            }
            for (int i = 0; i < files.length; i++) {
                if (files[i].isDirectory())
                    buildIndex(firstCharToTemporaryFile, firstCharToTemporaryFileStream, files[i]);
            }
        }
    }
    private void buildIndexFromFile(Map<String, File> firstCharToTemporaryFile, Map<String, DataOutputStream> firstCharToTemporaryFileStream, File file) throws Exception {
        Document doc = getHTMLDocument(file);
        
        Element[] elements = doc.getRootElements();
        for (int i = 0; i < elements.length; i++) {
            pickIndexes(firstCharToTemporaryFile, firstCharToTemporaryFileStream, file, elements[i]);
        }
    }
    private void pickIndexes(Map<String, File> firstCharToTemporaryFile, Map<String, DataOutputStream> firstCharToTemporaryFileStream, File file, Element element) throws IOException, BadLocationException {
        if (!element.isLeaf()) {
            int count = element.getElementCount();
            for (int j = 0; j < count; j++) {
                Element e = element.getElement(j);
                Tag tag = HTML.getTag(e.getName());
                if (tag == Tag.DT) {
                    String link = pickIndex(e);
                    int type = getType(link);
                    String title = getTitle(link);
                    if (title != null && title.length() > 0) {
                        String titleIgnoreCase = title.toUpperCase();
                        String index;
                        if (titleIgnoreCase.startsWith("SET") || titleIgnoreCase.startsWith("GET")) {
                            index = titleIgnoreCase.substring(0, Math.min(4, titleIgnoreCase.length()));
                        } else {
                            index = titleIgnoreCase.substring(0, Math.min(1, titleIgnoreCase.length()));
                        }
                        File temporaryFile = firstCharToTemporaryFile.get(index);
                        if (temporaryFile == null) {
                            temporaryFile = File.createTempFile("html2winhelp-", "." + index);
                            temporaryFile.deleteOnExit();
                            firstCharToTemporaryFile.put(index, temporaryFile);
                        }
                        DataOutputStream stream = firstCharToTemporaryFileStream.get(index);
                        if (stream == null) {
                            stream = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(temporaryFile, true)));
                            firstCharToTemporaryFileStream.put(index, stream);
                        }
                        stream.writeUTF(title);
                        stream.writeUTF(file.getParent().toString().replace('\\', '/') + ";" + link);
                    }
                } else {
                    pickIndexes(firstCharToTemporaryFile, firstCharToTemporaryFileStream, file, e);
                }
            }
        }
    }
    private String pickIndex(Element element) throws IOException, BadLocationException {
        String keyword = null;
        
        int count = element.getElementCount();
        for (int j = 0; j < count; j++) {
            Element e = element.getElement(j);
            AttributeSet attrSet = (AttributeSet) e.getAttributes().getAttribute(Tag.A);
            if (attrSet != null) {
                return (String) attrSet.getAttribute(Attribute.HREF);
            }
        }
        return null;
    }
    
    private String getTitle(String link) {
        int index = link.indexOf('#');
        if (index < 0) {
            int i = link.lastIndexOf('.');
            if (i >= 0) {
                int j = Math.max(link.lastIndexOf('.', i - 1), link.lastIndexOf('/', i - 1));
                return link.substring(j + 1, i);
            }
        } else {
            int j = link.indexOf('(', index + 1);
            if (j < 0) {
                return link.substring(index + 1);
            } else {
                int k = Math.max(index, link.lastIndexOf('.', j));
                return link.substring(k + 1, j);
            }
        }
        return null;
    }
    
    private String getInternalClassName(String link) {
        int i = link.indexOf('#');
        if (i < 0)
            i = link.length();
        i = link.lastIndexOf('.', i);   // skip .html
        
        int j = 0;
        while (j < i && link.charAt(j) == '.') {
            int k = link.indexOf('/', j);
            if (k < 0 || k + 1 >= i)
                break;
            j = k + 1;
        }
        
        return link.substring(j, i).replace('.', '$').replace('/', '.');
    }
    private String getCanonicalClassName(String link) {
        return getInternalClassName(link).replace('$', '.');
    }
    private String getClassName(String link) {
        String canonicalName = getCanonicalClassName(link);
        int index = canonicalName.lastIndexOf('.');
        return canonicalName.substring(index + 1);
    }
    
    private static final int TYPE = 0;
    private static final int CONSTRUCTOR = 1;
    private static final int METHOD = 2;
    private static final int FIELD = 3;
    private int getType(String link) {
        int index = link.indexOf('#');
        if (index < 0) {
            return TYPE;
        } else {
            int j = link.indexOf('(', index + 1);
            if (j < 0) {
                return FIELD;
            } else {
                String name = link.substring(index + 1, j);
                if (name.indexOf('.') >= 0 || name.equals(getClassName(link))) {
                    return CONSTRUCTOR;
                } else {
                    return METHOD;
                }
            }
        }
    }
    private boolean isNestedType(String link) {
        int index = link.indexOf('#');
        if (index >= 0)
            link = link.substring(0, index);
        index = link.lastIndexOf('.');
        if (index >= 0)
            link = link.substring(0, index);    // skip .html
        
        link = link.substring(link.lastIndexOf('/') + 1);
        index = link.lastIndexOf('.');
        return index >= 0;
    }
    
    private void outputIndex(String title, Collection<String> items) {
        Set<String> types = new TreeSet<String>();
        Set<String> constructors = new TreeSet<String>();
        Map<String, Map<String, Set<String>>> methods = new TreeMap<String, Map<String, Set<String>>>();
        Set<String> fields = new TreeSet<String>();
        
        for (String link : items) {
            int type = getType(link.substring(link.indexOf(';') + 1));
            switch (type) {
            default:
                assert false;
                // fall-through
            case TYPE:
                types.add(link);
                break;
            case CONSTRUCTOR:
                constructors.add(link);
                break;
            case METHOD:
                {
                    String key = link.substring(link.indexOf('#') + 1);
                    String compactKey = toSimpleName(key.substring(key.indexOf('(')));
                    Map<String, Set<String>> map = methods.get(compactKey);
                    if (map == null) {
                        map = new TreeMap<String, Set<String>>();
                        methods.put(compactKey, map);
                    }
                    Set<String> set = map.get(key);
                    if (set == null) {
                        set = new TreeSet<String>();
                        map.put(key, set);
                    }
                    set.add(link);
                }
                break;
            case FIELD:
                fields.add(link);
                break;
            }
        }
        
        for (String link : types) {
            String canonicalName = getCanonicalClassName(link.substring(link.indexOf(';') + 1));
            if (types.size() + fields.size() > 1) {
                outputIndex(title + " - " + canonicalName, link);
            } else {
                outputIndex(title, link);
            }
            
            if (!constructors.isEmpty()) {
                index.println("<UL>");
                Map<String, Set<String>> compaction = new TreeMap<String, Set<String>>();
                Iterator<String> it = constructors.iterator();
                while (it.hasNext()) {
                    String constructor = it.next();
                    if (getCanonicalClassName(constructor.substring(constructor.indexOf(';') + 1)).equals(canonicalName)) {
                        it.remove();
                        String key = constructor.substring(constructor.indexOf('#') + 1);
                        String compactKey = toSimpleName(key.substring(key.indexOf('(')));
                        Set<String> set = compaction.get(compactKey);
                        if (set == null) {
                            set = new TreeSet<String>();
                            compaction.put(compactKey, set);
                        }
                        set.add(constructor);
                    }
                }
                for (Map.Entry<String, Set<String>> entry : compaction.entrySet()) {
                    String key = entry.getKey();
                    Set<String> values = entry.getValue();
                    for (String constructor : values) {
                        if (values.size() > 1) {
                            key = constructor.substring(constructor.indexOf('#') + 1);
                            key = key.substring(key.indexOf('('));
                        }
                        outputIndex(key, constructor);
                    }
                }
                index.println("</UL>");
            }
        }
        
        for (Map<String, Set<String>> map : methods.values()) {
            for (Map.Entry<String, Set<String>> entry : map.entrySet()) {
                String key = entry.getKey();
                Set<String> set = entry.getValue();
                if (map.size() <= 1) {
                    key = key.substring(0, key.indexOf('(')) + toSimpleName(key.substring(key.indexOf('(')));
                    outputMethods(key, set);
                } else {
                    outputMethods(key, set);
                }
            }
        }
        
        Map<String, Set<String>> compaction = new TreeMap<String, Set<String>>();
        for (String link : fields) {
            String key = getClassName(link.substring(link.indexOf(';') + 1));
            Set<String> set = compaction.get(key);
            if (set == null) {
                set = new TreeSet<String>();
                compaction.put(key, set);
            }
            set.add(link);
        }
        for (Set<String> values : compaction.values()) {
            for (String link : values) {
                if (values.size() <= 1) {
                    outputIndex(title + " - " + getClassName(link.substring(link.indexOf(';') + 1)), link);
                } else {
                    outputIndex(title + " - " + getCanonicalClassName(link.substring(link.indexOf(';') + 1)), link);
                }
            }
        }
    }
    private String toSimpleName(String str) {
        StringTokenizer st = new StringTokenizer(str.substring(1, str.length() - 1), ",");
        StringBuilder result = new StringBuilder("(");
        while (st.hasMoreTokens()) {
            String s = st.nextToken().trim();
            boolean vararg = false;
            if (s.endsWith("...")) {
                vararg = true;
                s = s.substring(0, s.length() - 3);
            }
            s = s.substring(s.lastIndexOf('.') + 1);
            if (result.length() > 1)
                result.append(", ");
            result.append(s);
            if (vararg)
                result.append("...");
        }
        result.append(")");
        return result.toString();
    }
    private Set<String> loadFailedClassNames = new HashSet<String>();
    private void outputMethods(final String key, Set<String> methodLinks) {
        final Map<String, Set<String>> compaction = new TreeMap<String, Set<String>>();
        for (String link : methodLinks) {
            String compactKey = getClassName(link.substring(link.indexOf(';') + 1, link.indexOf('#')));
            compactKey = compactKey.substring(compactKey.lastIndexOf('.') + 1);
            Set<String> set = compaction.get(compactKey);
            if (set == null) {
                set = new TreeSet<String>();
                compaction.put(compactKey, set);
            }
            set.add(link);
        }
        
        class Method {
            String link;
            String className;
            Class<?> clazz;
            Set<Method> childMethods = new LinkedHashSet<Method>();
            public String toString() {
                return link;
            }
        }
        final Map<String, Method> classNameToMethod = new TreeMap<String, Method>();
        final Set<Method> allMethods = new LinkedHashSet<Method>();
        for (String link : methodLinks) {
            Method method = new Method();
            method.link = link;
            method.className = getInternalClassName(link.substring(link.indexOf(';') + 1));
            try {
                method.clazz = Class.forName(method.className);
            } catch (Throwable x) {
                boolean hold = (x instanceof ClassNotFoundException)
                            //|| (x instanceof VerifyError)           // TODO: remove
                            //|| (x instanceof NoClassDefFoundError)  // TODO: remove
                            || (x instanceof SecurityException)
                            ;
                Throwable cause = x.getCause();
                hold = hold || (cause instanceof ClassNotFoundException)
                            //|| (cause instanceof VerifyError)           // TODO: remove
                            //|| (cause instanceof NoClassDefFoundError)  // TODO: remove
                            || (cause instanceof SecurityException)
                            ;
                if (hold) {
                    if (loadFailedClassNames.add(method.className))
                        System.err.println("warning: failed to load " + method.className + ", caused by " + x.getClass().getName() + " (for " + method.link + ")");
                } else {
                    if (x instanceof Error)
                        throw (Error) x;
                    else
                        throw (RuntimeException) x;
                }
            }
            allMethods.add(method);
            classNameToMethod.put(method.className, method);
        }
        
        for (final Method method : allMethods) {
            new Runnable() {
                public void run() {
                    if (method.clazz == null)
                        return;
                    
                    findSuperClasses(method, method.clazz);
                }
                public void findSuperClasses(Method method, Class<?> clazz) {
                    Class<?> superClass = clazz.getSuperclass();
                    Class<?>[] interfaces = clazz.getInterfaces();
                    if (clazz.isInterface() && interfaces.length == 0) {
                        found(method, Object.class);
                    } else {
                        if (superClass != null)
                            found(method, superClass);
                        for (int i = 0; i < interfaces.length; i++) {
                            found(method, interfaces[i]);
                        }
                    }
                }
                public void found(Method method, Class<?> superType) {
                    Method m = classNameToMethod.get(superType.getName());
                    if (m != null) {
                        m.childMethods.add(method);
                    } else {
                        findSuperClasses(method, superType);
                    }
                }
            }.run();
        }
        
        for (Method m : new ArrayList<Method>(allMethods)) {
            allMethods.removeAll(m.childMethods);
        }
        
        final Set<Method> done = new HashSet<Method>();
        for (Method m : allMethods) {
            new Object() {
                public void output(Method method) {
                    if (done.add(method)) {
                        String className = method.className.replace('$', '.');
                        String simpleName = className.substring(className.lastIndexOf('.') + 1);
                        Set<String> set = compaction.get(simpleName);
                        if (set.size() <= 1)
                            className = simpleName;
                        
                        if (allMethods.contains(method)) {
                            outputIndex(key + " - " + className, method.link);
                        } else {
                            outputIndex("- " + className, method.link);
                        }
                        if (!method.childMethods.isEmpty()) {
                            index.println("<UL>");
                            for (Method m : method.childMethods) {
                                output(m);
                            }
                            index.println("</UL>");
                        }
                    }
                }
            }.output(m);
        }
        
    }
    private void outputIndex(String title, String link) {
        index.println("<LI><OBJECT type=\"text/sitemap\">");
        index.println("  <param name=\"Name\" value=\"" + toHTML(title) + "\">");
        link = link.substring(inputDir.toString().length() + 1);
        link = replaceAll(link.replace(';', '/'), " ", "");
        index.println("  <param name=\"Local\" value=\"" + toHTML(link) + "\">");
        index.println("</OBJECT>");
    }
    
    private void copyFiles(int level, File input, File output) throws IOException {
        System.out.println("converting " + output);
        output.mkdir();
        File[] files = input.listFiles();
        if (files != null) {
            for (int i = 0; i < files.length; i++) {
                if (!files[i].isDirectory())
                    copyFile(level, files[i], new File(output, files[i].getName()));
            }
            for (int i = 0; i < files.length; i++) {
                if (files[i].isDirectory())
                    copyFiles(level + 1, files[i], new File(output, files[i].getName()));
            }
        }
    }
    private void copyFile(int level, File input, File output) throws IOException {
        if (input.getName().endsWith(".html") || input.getName().endsWith(".htm")) {
            copyHTMLFile(level, input, output);
        } else {
            InputStream in = new FileInputStream(input);
            OutputStream out = new FileOutputStream(output);
            byte[] buffer = new byte[1024];
            for (;;) {
                int size = in.read(buffer);
                if (size < 0)
                    break;
                out.write(buffer, 0, size);
            }
            in.close();
            out.close();
        }
    }
    private void copyHTMLFile(int level, File input, File output) throws IOException {
        //System.out.println(input);
        
        CharSequence in = Decoder.decode(input, encodings);
        StringBuilder out;
        
        {
            Pattern pattern = Pattern.compile("(<A NAME=\")([^\"]*)(\">)");
            Matcher matcher = pattern.matcher(in);
            
            out = new StringBuilder(Math.max(10, 2 * in.length()));
            int index = 0;
            while (matcher.find()) {
                out.append(in.subSequence(index, matcher.end()));
                String name = in.subSequence(matcher.start(2), matcher.end(2)).toString();
                if (name.indexOf(' ') >= 0) {
                    out.append(in.subSequence(matcher.start(1), matcher.end(1)));
                    out.append(replaceAll(name, " ", ""));
                    out.append(in.subSequence(matcher.start(3), matcher.end(3)));
                }
                index = matcher.end();
            }
            out.append(in.subSequence(index, in.length()));
            for (int i = 0; i < out.length(); i++) {
                char ch = out.charAt(i);
                if (ch < ' ' /*|| 0x7F <= ch && ch <= 0xFF*/) {
                    if (!Character.isWhitespace(ch) && ch != 0xA5)
                        out.setCharAt(i, '?');
                }
            }
        }
        
        HTMLNormalizer normalizer = new HTMLNormalizer(level, input, out);
        Writer writer = new FileWriter(output);
        {
            String regex = "<META http-equiv=\"Content-Type\" content=\"text/html; charset=[^\"]*\">";
            String replacement = Pattern.compile(regex, Pattern.CASE_INSENSITIVE).matcher(normalizer.getResult().toString()).replaceAll("");
            writer.write(replacement);
        }
        writer.close();
    }
    
    private String toHTML(String str) {
        StringBuilder result = new StringBuilder();
        for (int i = 0; i < str.length();) {
            char ch = str.charAt(i++);
            if (ch == '&') {
                result.append("&amp;");
            } else if (ch == '<') {
                result.append("&lt;");
            } else if (ch == '>') {
                result.append("&gt;");
            } else {
                result.append(ch);
            }
        }
        return result.toString();
    }
    
    private HTMLDocument getHTMLDocument(File file) throws Exception {
        HTMLEditorKit kit = new HTMLEditorKit();
        HTMLDocument doc = (HTMLDocument) kit.createDefaultDocument();
        doc.putProperty("IgnoreCharsetDirective", Boolean.TRUE);
        CharSequence text = Decoder.decode(file, encodings);
        Reader reader = new StringReader(text.toString());  // Ummm...
        kit.read(reader, doc, 0);
        reader.close();
        return doc;
    }
    
    private void printTree(int indent, Element element) throws IOException, BadLocationException {
        for (int i = 0; i < indent; i++)
            System.out.print("  ");
        System.out.print(element);
        int start = element.getStartOffset();
        int end = element.getEndOffset();
        System.out.print("  ");
        System.out.println(element.getDocument().getText(start, end - start));
        
        int count = element.getElementCount();
        for (int j = 0; j < count; j++) {
            Element e = element.getElement(j);
            printTree(indent + 1, e);
        }
    }
    
    
    public static void main(String[] args) {
        try {
            if (args.length != 5) {
                System.out.println("usage: html2winhelp <short name> <title> <input dir> <output dir> <input-html-encodings>");
                System.exit(2);
            }
            new Main(args[0], args[1], new File(args[2]), new File(args[3]), args[4]);
            System.exit(0);
        } catch (Exception x) {
            x.printStackTrace();
            System.exit(1);
        }
    }
    
    
    /**
     * @deprecated
     */
    private static String replaceAll(String old, String regex, String replacement) {
        return Pattern.compile(regex).matcher(old).replaceAll(replacement);
    }
    
}

class Decoder {
    private static char[] buffer = new char[1024];
    public static CharSequence decode(File file, String encodings) throws IOException {
        StringBuilder in = new StringBuilder(Math.max(10, (int) file.length()));
        
        StringTokenizer st = new StringTokenizer(encodings, ",");
        do {
            String encoding = st.hasMoreTokens() ? st.nextToken().trim() : null;
            try {
                Reader reader = (encoding == null) ? new FileReader(file) : new FileReader(file, encoding);
                try {
                    for (;;) {
                        int size = reader.read(buffer);
                        if (size < 0)
                            break;
                        for (int i = 0; i < size; i++) {
                            int ch = buffer[i] & 0xFFFF;
                            switch (ch) {
                            default:
                                if (ch < ' ' || ch == 0x7F)
                                    throw new CharacterCodingException();
                                break;
                            case '\n':
                            case '\r':
                            case '\t':
                                break;
                            }
                        }
                        in.append(buffer, 0, size);
                    }
                } finally {
                    reader.close();
                }
                //System.err.println("encoding detected: " + file + " " + encoding);
                return in;
            } catch (CharacterCodingException x) {
            }
            in.delete(0, in.length());
        } while (st.hasMoreTokens());
        
        System.err.println("warning: encoding undetected " + file);
        Reader reader = new FileReader(file, "ISO8859_1");
        try {
            for (;;) {
                int size = reader.read(buffer);
                if (size < 0)
                    break;
                for (int i = 0; i < size; i++) {
                    int ch = buffer[i] & 0xFFFF;
                    switch (ch) {
                    default:
                        if (ch < ' ' || ch >= 0x7F) {
                            in.append('?');
                            break;
                        }
                        // fall-through
                    case '\n':
                    case '\r':
                    case '\t':
                        in.append((char) ch);
                        break;
                    }
                }
            }
        } finally {
            reader.close();
        }
        return in;
    }
}

class HTMLNormalizer {
    private final int level;
    private final File inputFile;
    private final CharSequence input;
    private final StringBuilder result;
    private int index;
    public HTMLNormalizer(int level, File inputFile, CharSequence input) {
        this.level = level;
        this.inputFile = inputFile;
        this.input = input;
        this.result = new StringBuilder(Math.max(10, input.length()));
        
        this.index = 0;
        outTagLoop();
    }
    
    private void outTagLoop() {
        while (index < input.length()) {
            char ch = input.charAt(index++);
            if (ch == '<') {
                char ch2 = (index < input.length()) ? input.charAt(index) : '\0';
                if (ch2 == '!') {
                    result.append(ch);
                    inCommentLoop();
                } else if ('A' <= ch2 && ch2 <= 'Z' || 'a' <= ch2 && ch2 <= 'z' || ch2 == '/') {
                    // in tag
                    result.append(ch);
                    inTagLoop();
                } else {
                    result.append("&lt;");
                }
            } else if (ch == '>') {
                result.append("&gt;");
            } else {
                result.append(ch);
            }
        }
    }
    
    private void inCommentLoop() {
        boolean inComment = false;
        while (index < input.length()) {
            char ch = input.charAt(index++);
            char ch2 = (index < input.length()) ? input.charAt(index) : '\0';
            if (ch == '-' && ch2 == '-') {
                result.append(ch2);
                index = index++;
                inComment = !inComment;
            } else {
                result.append(ch);
                if (!inComment && ch == '>')
                    return;
            }
        }
    }
    
    private void inTagLoop() {
        while (index < input.length()) {
            char ch = input.charAt(index);
            if (ch == '>') {
                result.append(ch);
                index++;
                break;
            }
            if (Character.isWhitespace(ch) || ch == '/') {
                result.append(ch);
                index++;
            } else {
                // TODO: href = "..." とか、スペースが入ってもいいらしい。
                if (index + 5 <= input.length() && input.subSequence(index, index + 5).toString().equalsIgnoreCase("HREF=")) {
                    href();
                    //attributeLoop();
                } else {
                    attributeLoop();
                }
            }
        }
    }
    
    private void href() {
        int start = index + 5;
        char quote = (start < input.length()) ? input.charAt(start) : '\0';
        if (quote == '\'' || quote == '\"') {
            start++;
        } else {
            quote = '\0';
        }
        
        int end = start;
        while (end < input.length()) {
            char ch = input.charAt(end);
            if (quote == '\0') {
                if (ch == '>' || Character.isWhitespace(ch) || ch == '/')
                    break;
            } else {
                if (ch == quote)
                    break;
            }
            end++;
        }
        
        String url = input.subSequence(start, end).toString();
        if (url.startsWith("..")) {
            int back = 0;
            Scanner scanner = new Scanner(url);
            scanner.useDelimiter("/");
            while (scanner.hasNext()) {
                String next = scanner.next();
                if (!next.equals(".."))
                    break;
                back++;
            }
            if (back > level) {
                System.err.println("warning: beyond-archive href=" + url + " in " + inputFile);
                index = end;
                if (quote != '\0')
                    index++;
                return;
            }
        }
        attributeLoop();
    }
    
    private void attributeLoop() {
        while (index < input.length()) {
            char ch = input.charAt(index);
            if (ch == '>' || Character.isWhitespace(ch) || ch == '/')
                return;
            
            result.append(ch);
            index++;
            if (ch == '=')
                break;
        }
        
        char quote = (index < input.length()) ? input.charAt(index) : '\0';
        if (quote == '\'' || quote == '\"') {
            result.append(quote);
            index++;
        } else {
            quote = '\0';
        }
        
        while (index < input.length()) {
            char ch = input.charAt(index);
            if (quote == '\0') {
                if (ch == '>' || Character.isWhitespace(ch) || ch == '/')
                    break;
            } else {
                if (ch == quote) {
                    result.append(ch);
                    index++;
                    break;
                }
            }
            
            if (ch == '>') {
                result.append("&gt;");
            } else if (ch == '<') {
                result.append("&lt;");
            } else if (ch == '\"') {
                result.append("&quot;");
            } else {
                result.append(ch);
            }
            index++;
        }
    }
    
    public CharSequence getResult() {
        return result;
    }
}

